var init = (function (Clutter, Gio, GLib, GObject, St, PanelMenu, PopupMenu) {
    'use strict';

    const ExtensionUtils = imports.misc.extensionUtils;
    const Main = imports.ui.main;
    const Me = ExtensionUtils.getCurrentExtension();
    // const GETTEXT_DOMAIN = 'my-indicator-extension'
    // const _ = ExtensionUtils.gettext
    const _ = (text) => text;
    const iconFrom = (filename) => Gio.icon_new_for_string([Me.path, 'icons', filename].join('/').replace(/\/+/igm, '/'));
    const UNIT = {
        celsius: "\u00b0C",
        fahrenheit: "\u00b0F",
        rpm: "RPM"
    };
    const ICON = {
        fan: { gicon: iconFrom('fan-symbolic.svg'), size: 12 },
        cpu: { gicon: iconFrom('cpu-symbolic.svg'), size: 12 },
        gpu: { gicon: iconFrom('gpu-symbolic.svg'), size: 14 },
        hdd: { gicon: iconFrom('hdd-symbolic.svg'), size: 14 },
        sensor: { gicon: iconFrom('sensor-symbolic.svg'), size: 14 }
    };
    class ConsoleUtil {
        constructor(program, ...args) {
            this._available = false;
            this._command = '';
            if (typeof program === 'string') {
                const path = GLib.find_program_in_path(program);
                this._available = Boolean(path);
                if (this._available) {
                    this._command = [path, ...args].join(' ');
                }
            }
            if (typeof program === 'object' && Array.isArray(program)) {
                this._available = program.every((file) => GLib.file_test(file, GLib.FileTest.EXISTS));
                if (this._available) {
                    this._command = ['cat', ...program, ...args].join(' ');
                }
            }
        }
        async execute(callback) {
            try {
                if (!this.available)
                    throw new Error("Util not available");
                const [ok, argv] = GLib.shell_parse_argv(this._command);
                if (!ok || !argv)
                    throw new Error("Unable to parse util arguments");
                const p = new Promise((resolve, reject) => {
                    let proc = Gio.Subprocess.new(argv, Gio.SubprocessFlags.STDOUT_PIPE | Gio.SubprocessFlags.STDERR_PIPE);
                    proc.communicate_utf8_async(null, null, (proc, res) => {
                        try {
                            if (!proc)
                                throw new Error("Util subprocess error");
                            let [, stdout, stderr] = proc?.communicate_utf8_finish(res);
                            if (!proc?.get_successful())
                                throw new Error(stderr);
                            resolve(stdout);
                        }
                        catch (e) {
                            reject(e);
                        }
                    });
                });
                callback(await p);
            }
            catch (e) {
                logError(e);
            }
        }
        get available() {
            return this._available;
        }
    }
    class LscpuUtil extends ConsoleUtil {
        constructor() {
            super('lscpu');
            if (this.available)
                this.update();
        }
        parse(str) {
            let name = str
                .split(/\n/)
                .find(row => row.includes("Model name:"))
                ?.replace("Model name:", '')
                .split('@')[0]
                .replace(/CPU/, ' ')
                .replace(/ +/igm, ' ')
                .trim();
            this._name = name || "Processor";
        }
        update() {
            return super.execute(this.parse.bind(this));
        }
        get name() {
            return this._name;
        }
    }
    class LsblkUtil extends ConsoleUtil {
        constructor() {
            super('lsblk', '-o', 'HCTL,MODEL,NAME,TRAN,VENDOR,TYPE', '-dnJ');
            if (this.available)
                this.update();
        }
        simplify(obj) {
            return obj.blockdevices.reduce((acc, { hctl, model }) => {
                if (!hctl)
                    return acc;
                const [a, b] = hctl.split(':');
                const key = ['drivetemp', 'scsi', a, b].join('-');
                acc[key] = model;
                return acc;
            }, {});
        }
        parse(str) {
            this._data = this.simplify(JSON.parse(str));
        }
        update() {
            return super.execute(this.parse.bind(this));
        }
        name(key) {
            return this._data[key];
        }
    }
    class SensorsUtil extends ConsoleUtil {
        constructor() {
            super('sensors', '-A', '-j');
            if (this.available) {
                this._lscpu = new LscpuUtil();
                this._lsblk = new LsblkUtil();
                this.update();
            }
        }
        simplify(obj) {
            const input = new RegExp(/input$/);
            const fans = new RegExp(/^fan/i);
            return Object.keys(obj).reduce((acc, key) => {
                const inner = Object.keys(obj[key]).find(k => input.test(k));
                if (inner) {
                    let value = obj[key][inner];
                    if (value > 0 || fans.test(key)) {
                        if (typeof value === 'number')
                            value = value.toString();
                        acc[key] = value;
                    }
                }
                else {
                    acc[key] = this.simplify(obj[key]);
                }
                return acc;
            }, {});
        }
        organize(data) {
            const cores = new RegExp(/^coretemp/i);
            const drives = new RegExp(/^drivetemp/i);
            const batteries = new RegExp(/^bat/i);
            const tpisa = new RegExp(/^thinkpad-isa/i);
            return Object.keys(data).reduce((acc, key) => {
                let value = data[key];
                if (Object.keys(value).length === 1) {
                    value = Object.values(value)[0];
                    if (typeof value === 'number') {
                        value = value.toString();
                    }
                }
                if (cores.test(key)) {
                    acc.cpu[this._lscpu.name] = value;
                    return acc;
                }
                if (drives.test(key)) {
                    acc.hdd[this._lsblk.name(key)] = value;
                    return acc;
                }
                if (batteries.test(key)) {
                    const k = key.split('-')[0];
                    acc.bat[k] = value;
                    return acc;
                }
                if (tpisa.test(key)) {
                    const fans = new RegExp(/^fan/i);
                    acc.fan = Object.keys(value)
                        .filter(k => fans.test(k))
                        .reduce((acc, k) => {
                        acc[k] = value[k].toString();
                        return acc;
                    }, {});
                    return acc;
                }
                acc.other[key] = value;
                return acc;
            }, { cpu: {}, hdd: {}, bat: {}, fan: {}, other: {} });
        }
        parse(str) {
            this._data = this.organize(this.simplify(JSON.parse(str)));
        }
        update() {
            return super.execute(this.parse.bind(this));
        }
        verify(obj) {
            return Object.keys(obj).length ? obj : false;
        }
        get cpu() {
            return this.verify(this._data.cpu);
        }
        get hdd() {
            return this.verify(this._data.hdd);
        }
        get bat() {
            return this.verify(this._data.bat);
        }
        get fan() {
            return this.verify(this._data.fan);
        }
        get other() {
            return this.verify(this._data.other);
        }
    }
    class IbmAcpiUtil extends ConsoleUtil {
        constructor() {
            super([
                '/proc/acpi/ibm/thermal',
                '/proc/acpi/ibm/fan'
            ]);
            this._data = {
                cpu: 0,
                gpu: 0,
                status: '...',
                speed: 0,
                level: '...',
                levels: []
            };
            // immediately get initial values
            if (this.available)
                this.update();
        }
        // temperatures: 43 50 0 0 0 0 0 0
        // status: enabled
        // speed: 2314
        // level: auto
        // commands: level<level>(<level>is 0 - 7, auto, disengaged, full - speed)
        // commands: enable, disable
        // commands: watchdog<timeout>(<timeout>is 0(off), 1 - 120(seconds))
        parse(str) {
            const getVal = (str) => str
                .replace(/\t+/, " ")
                .split(': ')[1];
            const row = str.split(/\n/);
            const [cpu, gpu] = getVal(row[0]).split(' ');
            let levels = [];
            if (row[4]) {
                const [mm, ...rest] = row[4]
                    .split('(<level> is')[1]
                    .replace(/ +/im, '')
                    .replace(/\)$/im, '')
                    .split(', ')
                    .filter(e => e !== 'disengaged');
                const max = parseInt(mm.split('-')[1]);
                const nums = Array.from(Array(max), (_, i) => (i + 1).toString());
                levels = [...nums, ...rest];
            }
            this._data = {
                cpu: parseInt(cpu),
                gpu: parseInt(gpu),
                status: getVal(row[1]),
                speed: parseInt(getVal(row[2])),
                level: getVal(row[3]),
                levels
            };
        }
        update() {
            return super.execute(this.parse.bind(this));
        }
        setLevel(next) {
            const cmd = `pkexec sh -c "echo level ${next} | tee /proc/acpi/ibm/fan"`;
            const [ok, argv] = GLib.shell_parse_argv(cmd);
            log(`Setting fan level to: ${next}`);
            if (ok && argv?.length) {
                try {
                    let proc = Gio.Subprocess.new(argv, Gio.SubprocessFlags.STDOUT_PIPE | Gio.SubprocessFlags.STDERR_PIPE);
                    proc.communicate_utf8_async(null, null, (proc, res) => {
                        if (proc) {
                            try {
                                let [, stdout, stderr] = proc?.communicate_utf8_finish(res);
                                if (!proc?.get_successful())
                                    throw new Error(stderr);
                                // done
                            }
                            catch (e) {
                                logError(e);
                            }
                        }
                    });
                }
                catch (e) {
                    logError(e);
                }
            }
        }
        get cpu() {
            return this._data.cpu.toString();
        }
        get gpu() {
            return this._data.gpu.toString();
        }
        get status() {
            return this._data.status;
        }
        get speed() {
            return this._data.speed.toString();
        }
        get level() {
            return this._data.level;
        }
        get levels() {
            return this._data.levels;
        }
    }
    const ThermalTitle = GObject.registerClass(class ThermalTitle extends PopupMenu.PopupBaseMenuItem {
        constructor(style_class, title) {
            super({
                style_class, reactive: false
            });
            this.setOrnament(PopupMenu.Ornament.HIDDEN);
            this._title = new St.Label({
                text: title,
                x_align: Clutter.ActorAlign.CENTER,
                x_expand: true
            });
            this.add_child(this._title);
        }
    });
    const ThermalItem = GObject.registerClass(class ThermalItem extends PopupMenu.PopupBaseMenuItem {
        constructor(style_class, updater, label, props) {
            super({ style_class, reactive: false });
            this._updater = updater;
            if (props?.icon || props?.hideOrnament) {
                this.setOrnament(PopupMenu.Ornament.HIDDEN);
            }
            if (props?.icon) {
                this.add_child(new St.Icon({
                    style_class: 'popup-menu-ornament',
                    icon_size: 14,
                    gicon: props?.icon?.gicon,
                    y_align: Clutter.ActorAlign.CENTER
                }));
            }
            this._label = new St.Label({
                style_class: 'label',
                text: label,
                x_align: Clutter.ActorAlign.START,
                x_expand: true
            });
            const value = updater();
            this._value = new St.Label({
                style_class: 'value',
                text: typeof value === 'string' ? value : 'N/A'
            });
            const els = [this._label, this._value];
            if (props?.unit) {
                this._unit = new St.Label({
                    style_class: 'unit',
                    text: props.unit
                });
                els.push(this._unit);
            }
            els.forEach(el => this.add_child(el));
        }
        update(unit, label) {
            const value = this._updater();
            this._value.set_text(value);
            if (unit)
                this._unit.set_text(unit);
            if (label)
                this._label.set_text(label);
        }
    });
    const ThermalGroup = GObject.registerClass(class ThermalGroup extends PopupMenu.PopupSubMenuMenuItem {
        constructor(style_class, updater, label, props) {
            super(label, true);
            this._children = [];
            this.add_style_class_name(style_class);
            this.setOrnament(PopupMenu.Ornament.HIDDEN);
            if (props?.icon)
                this.icon.gicon = props?.icon.gicon;
            const data = updater();
            this._children = Object.keys(data).map(key => ({
                key,
                // @ts-expect-error
                element: new ThermalItem("tpt-popup-submenu-item", () => updater()[key], key, { ...props, icon: false, hideOrnament: true })
            }));
            this._children.forEach(({ element }) => this.menu.addMenuItem(element, undefined));
        }
        update() {
            this._children.forEach(({ element }) => element.update());
        }
    });
    const ThermalDropDown = GObject.registerClass(class ThermalDropDown extends PopupMenu.PopupSubMenuMenuItem {
        constructor(style_class, updater, label, items, action) {
            super(label, false);
            this.add_style_class_name(style_class);
            this.setOrnament(PopupMenu.Ornament.HIDDEN);
            this._updater = updater;
            this._value = new St.Label({
                style_class: 'value',
                text: updater(),
                x_align: Clutter.ActorAlign.END,
                x_expand: false
            });
            this.insert_child_at_index(this._value, this.get_children().length - 1);
            this._children = items.map(key => {
                const element = new PopupMenu.PopupBaseMenuItem({
                    style_class: 'tpt-popup-dropdown-item'
                });
                element.add_child(new St.Label({
                    style_class: 'label',
                    text: key,
                    x_align: Clutter.ActorAlign.END,
                    x_expand: true
                }));
                element.connect("activate", () => action(key));
                return ({ key, element });
            });
            this._children.forEach(({ element }) => this.menu.addMenuItem(element, undefined));
        }
        update() {
            this._value.set_text(this._updater());
        }
    });
    const IndicatorItem = GObject.registerClass(class IndicatorItem extends St.BoxLayout {
        constructor(style_class, updater, icon, unit) {
            super({ style_class });
            this._updater = updater;
            this._icon = new St.Icon({
                gicon: icon.gicon,
                icon_size: icon.size,
                style_class: 'icon',
                y_align: Clutter.ActorAlign.CENTER
            });
            this._value = new St.Label({
                text: updater(),
                style_class: 'value',
                y_align: Clutter.ActorAlign.CENTER
            });
            this._unit = new St.Label({
                text: unit,
                style_class: 'unit',
                y_align: Clutter.ActorAlign.CENTER
            });
            this._off = unit === 'RPM' ? new St.Label({
                text: 'OFF',
                style_class: 'unit',
                y_align: Clutter.ActorAlign.CENTER,
            }) : false;
            const els = [this._icon, this._value, this._unit];
            els.forEach(el => this.add_child(el));
            if (this._off)
                this.add_child(this._off);
            this.update();
        }
        off(hide) {
            if (!this._off)
                return;
            this._off.visible = hide;
            this._unit.visible = !hide;
            this._value.visible = !hide;
        }
        update(unit) {
            const v = this._updater();
            this._value.set_text(v);
            if (unit)
                this._unit.set_text(unit);
            this.off(v === "0");
        }
    });
    const Indicator = GObject.registerClass(class Indicator extends PanelMenu.Button {
        constructor(...args) {
            super(...args);
            // ui
            this._buttonLayout = new St.BoxLayout({ style_class: 'tpt-button-layout' });
            // internals
            this._bindings = [];
            this._elements = ['cpu', 'gpu', 'speed'];
            this._checkElements = () => {
                // check the reading for dedicated gpu
                // remove the element in case of anything <= 0
                // since reading of unavailable sensor appears to be either -128 or 0
                if (parseInt(this._tpAcpi.gpu) <= 0) {
                    this._elements = this._elements.filter(e => e !== 'gpu');
                }
            };
            this._start = async () => {
                if (this._tpAcpi.available) {
                    await this._tpAcpi.update();
                    this._checkElements();
                    this.appendButton();
                    if (this._sensors.available)
                        await this._sensors.update();
                    this.appendPopupMenu();
                }
                else {
                    this._buttonLayout.add_child(new St.Label({ text: "ThinkPad ACPI is not available" }));
                }
                this.add_actor(this._buttonLayout);
                if (this._tpAcpi.available) {
                    this._updateInterval = setInterval(this._update.bind(this), 5000);
                    this.connect("destroy", this._destroy.bind(this));
                }
            };
            // utility fn to convert celsius to fahrenheit
            this._toFahrenheit = (temp) => (temp * (9 / 5)) + 32;
            // perform update
            this._update = async () => {
                // get the data
                await Promise.all([
                    this._tpAcpi.available ? this._tpAcpi.update() : null,
                    this._sensors.available ? this._sensors.update() : null
                ].filter(p => p != null));
                // update values for non static bound elements
                this._bindings
                    .filter(({ type }) => !['separator', 'tpt-popup-title'].includes(type))
                    .forEach(({ element }) => {
                    if (typeof element.update === 'function')
                        element.update();
                });
            };
            this._tpAcpi = new IbmAcpiUtil();
            this._sensors = new SensorsUtil();
            this._start();
            // applying the style class overrides default styling
            // this.style_class = "tpt"
        }
        // create and attach element to bindings
        _attach(Element, type, ...args) {
            if (type === 'separator') {
                this.menu.addMenuItem(new PopupMenu.PopupSeparatorMenuItem);
                return;
            }
            const element = new Element(type, ...args);
            this._bindings.push({ type, element });
            if (['tpt-button'].includes(type))
                this._buttonLayout.add_child(element);
            if (['tpt-popup-dropdown', 'tpt-popup-submenu', 'tpt-popup-title', 'tpt-popup-item'].includes(type))
                this.menu.addMenuItem(element);
        }
        // cleanup on destroy
        _destroy() {
            if (this._updateInterval?.get_id()) {
                this._updateInterval.destroy();
            }
        }
        // is element enabled?
        element(key) {
            return this._elements.includes(key);
        }
        appendButton() {
            // attach cpu, gpu, speed indicator
            if (this.element('cpu'))
                this._attach(IndicatorItem, "tpt-button", () => this._tpAcpi.cpu, ICON.cpu, UNIT.celsius);
            if (this.element('gpu'))
                this._attach(IndicatorItem, "tpt-button", () => this._tpAcpi.gpu, ICON.gpu, UNIT.celsius);
            if (this.element('speed'))
                this._attach(IndicatorItem, "tpt-button", () => this._tpAcpi.speed, ICON.fan, UNIT.rpm);
        }
        appendPopupMenu() {
            // Sensors
            if (this._sensors.available) {
                // attach submenu for every cpu
                if (this._sensors.cpu)
                    Object.keys(this._sensors.cpu)
                        .forEach(k => this._attach(ThermalGroup, "tpt-popup-submenu", () => this._sensors.cpu[k], k, { icon: ICON.cpu, unit: UNIT.celsius }));
                // attach submenu for disks
                if (this._sensors.hdd)
                    this._attach(ThermalGroup, "tpt-popup-submenu", () => this._sensors.hdd, "Disks", { icon: ICON.hdd, unit: UNIT.celsius });
                // attach submenu for other thermal sensors
                if (this._sensors.other)
                    this._attach(ThermalGroup, "tpt-popup-submenu", () => this._sensors.other, "Thermal", { icon: ICON.sensor, unit: UNIT.celsius });
                // attach submenu for fans
                if (this._sensors.fan)
                    this._attach(ThermalGroup, "tpt-popup-submenu", () => this._sensors.fan, "Cooling", { icon: ICON.fan, unit: UNIT.rpm });
            }
            // tpAcpi groups
            if (this._tpAcpi.available) {
                // add ACPI items to popup
                this._attach(ThermalTitle, "tpt-popup-title", "ACPI");
                if (this.element('cpu')) {
                    this._attach(ThermalItem, "tpt-popup-item", () => this._tpAcpi.cpu, "CPU", { unit: UNIT.celsius, icon: ICON.cpu });
                }
                if (this.element('gpu')) {
                    this._attach(ThermalItem, "tpt-popup-item", () => this._tpAcpi.gpu, "GPU", { unit: UNIT.celsius, icon: ICON.gpu });
                }
                // add Fan control items to popup
                this._attach(ThermalTitle, "tpt-popup-title", "Fan control");
                this._attach(ThermalItem, "tpt-popup-item", () => this._tpAcpi.status, "Status", { hideOrnament: true });
                this._attach(ThermalItem, "tpt-popup-item", () => this._tpAcpi.speed, "Speed", { unit: UNIT.rpm, hideOrnament: true });
                // show dropdown if fan control is enabled
                if (this._tpAcpi.status === 'enabled' && this._tpAcpi.levels.length) {
                    this._attach(ThermalDropDown, "tpt-popup-dropdown", () => this._tpAcpi.level, "Levels", this._tpAcpi.levels, a => this._tpAcpi.setLevel(a));
                }
                else {
                    this._attach(ThermalItem, "tpt-popup-item", () => this._tpAcpi.level, "Level", { hideOrnament: true });
                }
            }
        }
    });
    class ThinkPadThermal {
        constructor(uuid) {
            this._uuid = uuid;
        }
        enable() {
            this._indicator = new Indicator(0.5, _('ThinkPad Thermal'));
            Main.panel.addToStatusArea(this._uuid, this._indicator);
        }
        disable() {
            this._indicator.destroy();
            this._indicator = null;
        }
    }
    function extension (meta) {
        return new ThinkPadThermal(meta.uuid);
    }

    return extension;

})(imports.gi.Clutter, imports.gi.Gio, imports.gi.GLib, imports.gi.GObject, imports.gi.St, imports.ui.panelMenu, imports.ui.popupMenu);
